/*
** Module   : MenuItemConfig.java
** Abstract : Menu item widget attribute container.
**
** Copyright (c) 2014-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 VIG 20141118          Created initial version. 
*/
package com.goldencode.p2j.ui;

import java.io.*;

/**
 * MENU-ITEM widget configuration container.
 */
public class MenuItemConfig
extends BaseConfig
{
   /**
    * Default SUBTYPE value. Means that this menu-item is normal element of menu
    * with label, sub-menu or some action, that called when this menu-item is chosen.
    */
   public static final String SBTP_NORMAL = "NORMAL";
   
   /**
    * This SUBTYPE denotes, that a blank line is inserted at this point in the menu.
    */
   public static final String SBTP_SKIP = "SKIP";
   
   /**
    * This SUBTYPE denotes, that a rule or line is inserted at this point in the menu.
    */
   public static final String SBTP_RULE = "RULE";
   
   /**
    * Parent widget id.
    */
   public int parentId;
   
   /**
    * Hot key for menu calling. Set by adding '&' to label.
    * The key is the first character after '&' or first 
    * character of the whole label if there is no '&' in it.
    */
   public Character mnemonic;
   
   /**
    * READ-ONLY attr. Specifies that this menu item is read-only text. 
    * The user cannot choose this item.
    */
   public boolean readOnly = false;
   
   /**
    * TOGGLE-BOX attr. Specifies that the menu item is displayed as a checkbox 
    * that the user can toggle on or off. In environments that do not support 
    * this option, it is ignored.
    */
   public boolean toggleBox = false;
   
   /**
    * ACCELERATOR attr. Specifies a keyboard accelerator for this menu item. 
    * The value keylabel must be character-string expression that
    * evaluates to a valid key label recognized by Progress, such as a, F1, or ALT-SHIFT-F1.
    */
   public String accelerator;
   
   /**
    * SUBTYPE attr. Can have one of three following values: {@link MenuItemConfig#SBTP_NORMAL},  
    * {@link MenuItemConfig#SBTP_SKIP} or {@link MenuItemConfig#SBTP_RULE}. Specifies an appearance
    * and behavior of this menu-item. Can be specified only before realization. 
    */
   public String subtype = SBTP_NORMAL;

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      MenuItemConfig mcfg = (MenuItemConfig) config;
      
      parentId = mcfg.parentId;
      subtype = mcfg.subtype;
      mnemonic = mcfg.mnemonic;
      readOnly = mcfg.readOnly;
      toggleBox = mcfg.toggleBox;
      accelerator = mcfg.accelerator;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      
      parentId = in.readInt();
      subtype = in.readLine();
      mnemonic = in.readChar();
      readOnly = in.readBoolean();
      toggleBox = in.readBoolean();
      accelerator = in.readLine();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      
      out.writeInt(parentId);
      out.writeChars(subtype);
      out.writeChar(mnemonic);
      out.writeBoolean(readOnly);
      out.writeBoolean(toggleBox);
      out.writeChars(accelerator);
   }
}
