/*
** Module   : MenuItemWidget.java
** Abstract : server side menu item widget implementation
**
** Copyright (c) 2014-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 VIG 20141118          Created initial version.
*/
package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * This class represents menu-item widget.
 */
@LegacyResource(resource = LegacyResource.MENU_ITEM)
public class MenuItemWidget
extends GenericWidget<MenuItemConfig>
implements SubTypeAttribute
{
   /**
    * Default constructor for dynamic menu-items  
    */
   public MenuItemWidget(boolean dynamic)
   {
      super(dynamic, new MenuItemConfig());
   }

   /**
    * Default constructor for static menu-item
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    label
    *           MENU-ITEM label.          
    */
   public MenuItemWidget(String name, String label)
   {
      super(false, new MenuItemConfig());
      name(name);
      setLabel(label);
   }

   /**
    * constructor for static menu-item of non standard sub-type: RULE or SKIP.
    * 
    * @param    subtype
    *           MENU-ITEM sub-type. One of 3 constants: {@link MenuItemConfig#SBTP_NORMAL},
    *           {@link MenuItemConfig#SBTP_RULE} or {@link MenuItemConfig#SBTP_SKIP}                    
    */
   public MenuItemWidget(String subtype)
   {
      super(false, new MenuItemConfig());
      setSubType(subtype);
   }
   
   /**
    * Sets the label text, parsing mnemonic.
    * 
    * TODO: find out what is going on with undocumented feature of very long label
    * See parent's javadoc.
    * 
    * @param    label
    *           The new label text.
    */
   @Override
   public void setLabel(String label)
   {
      config.setDynamicLabel(label);
      
      int idx = label.indexOf('&');
      
      // TODO find out what mnemonic will be if '&' is the last symbol.
      if (idx < 0 || idx == label.length() - 1) 
      {
         config.mnemonic = label.charAt(0);
      }
      else
      {
         config.mnemonic = label.charAt(idx + 1);
      }
      
      // TODO It seems that pushScreenDefinition is not needed here 
      //      because the menu-item is not in a frame. Not sure because
      //      a popup menu can be owned by widget, for example, by button.
   }

   /**
    * Get MENU-ITEM mnemonic key.
    * 
    * @return  Mnemonic character.
    *          At some moment of dynamic menu-item construction it can be null. 
    */
   public Character getMnemonic()
   {
      return config.mnemonic;
   }
   
   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   @Override
   public logical isReadOnly()
   {
      return new logical(config.readOnly);
   }

   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   @Override
   public void setReadOnly(logical r)
   {
      if (r != null)
         setReadOnly(r.booleanValue());
   }
   
   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   public void setReadOnly(boolean r)
   {
      config.readOnly = r;
   }

   /**
    * Gets the TOGGLE-BOX attribute value.
    * 
    * @return  See above.
    */
   public boolean isToggleBox()
   {
      return config.toggleBox;
   }

   /**
    * Sets the TOGGLE-BOX attribute of the menu-item.
    * 
    * @param    toggleBox
    *           <code>true</code> if the menu-item should have a checkbox.
    */
   public void setToggleBox(boolean toggleBox)
   {
      config.toggleBox = toggleBox;
   }
   
   /**
    * Gets the display color writable attribute.
    *
    * @return   The current value of the display color attribute.
    */
   @Override
   public integer getDColor()
   {
      int color = config.dcolor.color();
      return (color == -1) ? new integer() : new integer(color);
   }
   
   /**
    * Sets the display color.
    *
    * @param    param
    *           The display color.
    */
   public void setDColor(NumberType param)
   {
      config.dcolor = new ColorSpec(param.intValue()).convert();
   }

   /**
    * Sets the display color.
    * 
    * @param    displayColor
    *           The display color.
    */
   public void setDColor(Color displayColor)
   {
      config.dcolor = displayColor;
   }

   /**
    * Sets the display color.
    * 
    * @param    displayColor
    *           The display color.
    */
   public void setDColor(int displayColor)
   {
      config.dcolor = new ColorSpec(displayColor).convert();
   }

   /**
    * Get foreground color.
    * 
    * @return   See above.
    */
   @Override
   public integer getFgColor()
   {
      return (config.fgcolor == -1) ? new integer() : new integer(config.fgcolor);
   }

   /**
    * Set foreground color.
    * 
    * @param   fgcolor
    *          Color number.
    */
   @Override
   public void setFgColor(integer fgcolor)
   {
      config.fgcolor = fgcolor.intValue();
   }

   /**
    * Set foreground color.
    * 
    * @param   fgcolor
    *          Color number.
    */
   public void setFgColor(int fgcolor)
   {
      config.fgcolor = fgcolor;
   }

   /**
    * Get background color.
    * 
    * @return   See above.
    */
   @Override
   public integer getBgColor()
   {
      return (config.bgcolor == -1) ? new integer() : new integer(config.bgcolor);
   }

   /**
    * Set background color.
    * 
    * @param   bgcolor
    *          Color number.
    */
   @Override
   public void setBgColor(integer bgcolor)
   {
      config.bgcolor = bgcolor.intValue();
   }

   /**
    * Set background color.
    * 
    * @param   bgcolor
    *          Color number.
    */
   public void setBgColor(int bgcolor)
   {
      config.bgcolor = bgcolor;
   }

   /**
    * Gets the input (prompt/edit mode) color.
    *
    * @return   The current value of the display color attribute.
    */
   @Override
   public integer getPfColor()
   {
      int color = config.pfcolor.color();
      return (color == -1) ? new integer() : new integer(color);
   }
   
   /**
    * Sets the input (prompt/edit mode) color.
    *
    * @param    param
    *           The display color.
    */
   public void setPfColor(NumberType param)
   {
      config.pfcolor = new ColorSpec(param.intValue()).convert();
   }

   /**
    * Sets the input (prompt/edit mode) color.
    * 
    * @param    displayColor
    *           The display color.
    */
   public void setPfColor(Color displayColor)
   {
      config.pfcolor = displayColor;
   }

   /**
    * Sets the input (prompt/edit mode) color.
    * 
    * @param    displayColor
    *           The display color.
    */
   public void setPfColor(int displayColor)
   {
      config.pfcolor = new ColorSpec(displayColor).convert();
   }

   /**
    * Gets this menu-item accelerator.
    * 
    * @return  See above.
    */
   public String getAccelerator()
   {
      return config.accelerator;
   }

   /**
    * Sets this menu-item accelerator.
    * 
    * @param  accelerator.
    *         The valid keylabel, that will be target accelerator.
    */
   public void setAccelerator(String accelerator)
   {
      config.accelerator = accelerator;
   }

   /**
    * Gets this menu-item parent (MENU or SUBMENU).
    * 
    * @return  See above.
    */
   @Override
   public GenericWidget<?> getParent()
   {
      return LogicalTerminal.getWidgetForId(config.parentId);
   }

   /**
    * Sets this menu-item parent (MENU or SUBMENU).
    * 
    * @param  parent.
    *         MENU or SUBMENU widget.
    */
   @Override
   public void setParent(GenericWidget<?> parent)
   {
      if (parent instanceof MenuWidget || parent instanceof SubMenuWidget)
      {
         config.parentId = parent.getId();
      }
      else
      {
         // TODO "**Unable to assign UNKNOWN value to attribute PARENT on MENU-ITEM widget. (4083)"
         String msgTemplate = "**%s widget cannot be a parent of a MENU-ITEM widget";
         ErrorManager.recordOrThrowError(4075, String.format(msgTemplate, parent.type()), false);
      }
   }

   /**
    * Returns the sub-type of this menu-item.
    * 
    * @return   See above.
    */
   @Override
   @LegacyAttribute(name = "SUBTYPE")
   public character getSubType()
   {
      if (!_dynamic())
      {
         String[] msg = new String[]{"**SUBTYPE is not a queryable attribute for MENU m"};
         ErrorManager.recordOrShowError(new int[]{4052}, msg, false, false, false);
      }
      
      return new character(config.subtype);
   }
   
   /**
    * Sets the the sub-type of the object.
    * 
    * @param    value
    *           The new value of the sub-type object attribute.
    */
   @LegacyAttribute(name = "SUBTYPE", setter = true)
   public void setSubType(character value)
   {
      if (value == null || value.isUnknown())
      {
         String templ = "**Unable to assign UNKNOWN value to attribute SUBTYPE on MENU-ITEM %s.";
         String[] msg = new String[] { String.format(templ, name()) };
         ErrorManager.recordOrShowError(new int[] {4083}, msg, false, false, false);
      }
      else
      {
         setSubType(value.getValue());
      }
   }
   
   /**
    * Sets the the sub-type of the object.
    * 
    * @param    value
    *           The new value of the sub-type object attribute.
    */
   @LegacyAttribute(name = "SUBTYPE", setter = true)
   public void setSubType(String value)
   {
      if (!_dynamic())
      {
         String templ = "**Unable to set attribute SUBTYPE for MENU-ITEM %s";
         String[] msg = new String[] { String.format(templ, name()) };
         ErrorManager.recordOrShowError(new int[] {4078}, msg, false, false, false);
      }
      
      if (getParent() != null)
      {
         String msg = "**Unable to set SUBTYPE because the MENU-ITEM widget has been realised";
         ErrorManager.recordOrShowError(new int[] {4053}, new String[] {msg}, false, false, false);
      }
      
      config.subtype = value;
   }

   /**
    * Finds the first MENU-ITEM with specified name in all menus.
    * 
    * @param   name
    *          The name of searched MENU-ITEM
    *          
    * @return  The first found MenuItemWidget or null.
    */
   public static MenuItemWidget findMenuItemStatic(String name)
   {
      // TODO implement
      return null;
   }
   
}