/*
** Module   : SharedVariableManager.java
** Abstract : provides context-local scoped resource pools.
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050729   @21924 Provides context-local scoped variable 
**                           and stream pools.
** 002 GES 20050808   @22069 Moved ContextLocal inner class to a stand-
**                           alone class to promote reuse.
** 003 GES 20050818   @22163 Implemented the Scopeable interface to allow
**                           this class to be automatically notified of
**                           all scope start and end occurrences. This
**                           removes the need for client code to be
**                           instrumented to provide this service. The
**                           use of this class is now tied to the
**                           TransactionManager.
** 004 ECF 20050920   @22887 Implement ScopeableFactory interface. Allows
**                           generic access to singleton instance of this
**                           class from TransactionManager.
** 005 GES 20051004   @22930 Added support for shared buffers and error
**                           manager processing.
** 006 GES 20051101   @23163 Added support for shared frames.
** 007 GES 20060123   @24014 Converted from thread local to context local
**                           and moved initialization code to be based
**                           on the standard directory-based server
**                           initialization routines.
** 008 GES 20061010   @30262 Removed global option from frame support
**                           since this is not allowed in Progress.
** 009 ECF 20070328   @32652 Added shared temp-table support. Added
**                           context container for temp tables and
**                           add/lookup methods for temp tables.
** 010 SIY 20070711   @34430 Changed addFrame() to return existing frame
**                           if one is present in storage.
** 011 GES 20080822   @39520 Slight modification proposed by SVL to allow
**                           the currently registered global var of the
**                           same name to be returned as a result of the
**                           addVariable() method.
** 012 SIY 20080904   @39718 Inferred generics.
** 013 GES 20080918   @39857 Major rewrite to handle scopes differently.
**                           Now each resource supports some or all of
**                           the following: CURRENT, NEXT, GLOBAL. NEXT
**                           is the new feature that allows a resource
**                           add to wait until the next scope is opened
**                           before being added to the dictionary. Add'l
**                           signatures were added to allow a default
**                           scope that differs by resource type.
** 014 GES 20090422   @41914 Converted to standard string formatting.
** 015 CA  20131013          Added no-op scopeDeleted() method, required by the changes in
**                           Scopeable interface.
** 016 OM  20141203          Added double-initialization check for NEW SHARED GLOBAL EXTENT-s.
** 017 VIG 20141215          Added methods for shared menus support.
*/

package com.goldencode.p2j.util;

import java.util.*;

import com.goldencode.p2j.persist.Temporary;
import com.goldencode.p2j.security.ContextLocal;

/**
 * Provides context-local scoped variable, stream, buffer, frame and temp
 * table pools.
 * <p>
 * All references can be added to the top scope (closest enclosing scope) or
 * to the global scope.  Lookups return the topmost reference by the given
 * name or <code>null</code> if no reference by that name exists.  The backing
 * function is provided by the {@link ScopedSymbolDictionary} and this class
 * exposes the necessary helpers as static methods which de-reference the
 * pool that is localized to the context that is specific to the current
 * authenticated user.  Please see <code>SecurityManager</code> for more
 * details on this facility.
 * <p>
 * The resource dictionaries (pools) are separate since they require
 * different namespaces.  Where possible, the returned reference is of
 * a specific type, but at this time for all non-stream and non-temp table
 * resources, only <code>Object</code> references can be returned.
 * <p>
 * Buffers are persistent object representations of relational database
 * records.
 * <p>
 * Temp tables are actually buffers which point to temp tables, but this is
 * simply an implementation detail.  They represent shared temp table
 * definitions, as opposed to shared buffers.  They are managed separate from
 * regular buffers to ensure a separate namespace is available.
 * <p>
 * Frames are object representations of a screen buffer that is specific to
 * a given user-defined visual layout.
 * <p>
 * Implements the {@link Scopeable} interface to allow this class to be
 * automatically notified of all scope start and end occurrences. This removes
 * the need for client code to be instrumented to provide this service. The
 * use of this class is now tied to the {@link TransactionManager} class. In
 * particular, the singleton instance of this class is used in each context-
 * local list of {@link Scopeable} objects as a dispatcher. This bit
 * of nastiness is necessary since the Java Interface can't be implemented
 * on a static basis but instead requires an instance method implementation.
 * These instance methods then call the static counterparts which is where
 * the content-local versions of the resource pools are accessed.
 * <p>
 * The initialization processing is implemented in the {@link #initialize}
 * method and it is called during server startup based on an entry in the
 * directory until the path "/server/&lt;serverId&gt;/startup" or in
 * "/server/default/startup".
 *
 * @author    GES
 */
public class SharedVariableManager
implements Scopeable,
           ScopeableFactory
{
   /** Missing variable error message. */
   private static final String VAR_ERR_MSG = "Shared variable %s has not yet been created";
   
   /** Missing stream error message. */
   private static final String STREAM_ERR_MSG = "Shared stream %s not defined yet";
   
   /** Missing buffer error message. */
   private static final String BUF_ERR_MSG = "Unable to find shared buffer for %s";
   
   /** Missing temp-table buffer error message. */
   private static final String TT_ERR_MSG = "Unable to locate shared temp-table or work-table " +
                                            "definition for table %s in procedure";
   
   /** Missing frame error message. */
   private static final String FR_ERR_MSG = "Shared frame %s not created yet";
   
   /** Variable pool which is context-local. Initializes upon first use. */ 
   private static ContextContainer<Object> vars = new ContextContainer<>();
   
   /** 
    * Global shared extent variable set that were 'added' but not yet initialized.
    * Programmer must check if wasInitialized before re-initializing the array.
    */ 
   private static ContextLocal<HashSet<BaseDataType[]>> uninitializedGlobalExtents = 
      new ContextLocal()
      {
         /** 
          * Create the HashSet of global extents variables for this context.
          */
         @Override
         protected Object initialValue()
         {
            return new HashSet<BaseDataType[]>();
         }
      };
   
   /** Pool of streams which is context-local. Initializes upon first use. */ 
   private static ContextContainer<Object> streams = new ContextContainer<>();
   
   /** Pool of buffers which is context-local. Initializes upon first use. */ 
   private static ContextContainer<Object> buffers = new ContextContainer<>();
   
   /** Temp-table pool which is context-local. Initializes upon first use. */ 
   private static ContextContainer<Object> tempTables = new ContextContainer<>();
   
   /** Pool of frames which is context-local. Initializes upon first use. */ 
   private static ContextContainer<Object> frames = new ContextContainer<>();
   
   /** Pool of frames which is context-local. Initializes upon first use. */ 
   private static ContextContainer<Object> menus = new ContextContainer<>();
   
   /** List of managed resources. */
   private static ContextContainer[] resources = 
   {
      vars,
      streams,
      buffers,
      tempTables,
      frames,
      menus
   };
   
   /** Singleton instance of this class (across entire JVM, all contexts). */ 
   private static SharedVariableManager single = null;
   
   /** Flag to detect and avoid attempts to initialize more than once. */
   private static boolean registered = false;
   
   /**
    * Implements the singleton nature of this class, constructs this instance
    * if it hasn't already been constructed.
    *
    * @return   The single instance of this class.
    */
   public static synchronized SharedVariableManager getInstance()
   {
      if (single == null)
      {
         single = new SharedVariableManager();
      }
      
      return single;
   }
   
   /**
    * Initialize the shared variable framework. This registers with the
    * {@link TransactionManager TransactionManager} a factory object which
    * creates a {@link Scopeable} instance.
    *
    * @throws  IllegalStateException
    *          If invoked more than once.
    */
   public static synchronized void initialize()
   {
      if (!registered)
      {
         TransactionManager.registerScopeableFactory(getInstance());
         registered = true;
      }
      else
      {
         throw new IllegalStateException("Initialize called > 1 time!");
      }
   }
   
   /**
    * Creates a new scope for all resource types.
    */
   public static void addScope()
   {
      for (ContextContainer<Object> cc : resources)
      {
         WorkArea<Object> wa = cc.obtainWA();
         wa.shared.addScope(null);
         
         if (wa.pending != null)
         {
            Set<String> keys = wa.pending.keySet();
            
            for (String next : keys)
            {
               wa.shared.addSymbol(false, next, wa.pending.get(next));
            }
            
            wa.pending = null;
         }
      }
   }
   
   /**
    * Deletes the top-most scope for all resource types.
    */
   public static void deleteScope()
   {
      for (ContextContainer<Object> cc : resources)
      {
         WorkArea<Object> wa = cc.obtainWA();
         wa.shared.deleteScope();
      }
   }
   
   /**
    * Adds a variable to the next scope that is opened.
    *
    * @param    <T> 
    *           Variable type.
    * @param    name
    *           The variable name.
    * @param    var
    *           The variable reference.
    *
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(String name, T var)
   {
      return addVariable(ScopeLevel.NEXT, name, var);
   }
   
   /**
    * Adds a variable to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is
    * opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    * <p>
    * Any attempt to add second global variable of the same name, will cause
    * the passed-in instance to be ignored (the currently registered instance
    * will not be replaced).  In addition, the currently registered instance
    * will be returned.
    *
    * @param    <T> 
    *           Variable type.
    * @param    scope
    *           Specifies the target scope.
    * @param    name
    *           The variable name.
    * @param    var
    *           The variable reference.
    *
    * @return   The <code>var</code> instance given if it is successfully
    *           added, otherwise the currently registered global instance
    *           with the same name.
    */
   public static <T> T addVariable(ScopeLevel scope, String name, T var)
   {
      WorkArea<Object> wa = vars.obtainWA();
      
      if (scope == ScopeLevel.GLOBAL)
      {
         T current = (T) wa.shared.getSymbolAtScope(name, -1);
         
         // ignore repeated global definitions
         if (current != null)
         {
            return current;
         }
      }
      
      addWorker(wa, scope, name, var);
      
      // for new global shared extent declared first time we prepare for initialization
      // subsequent request MUST not be re-initialized. If the variable had already been
      // declared/initialized then this method should have returned with found instance
      if (scope == ScopeLevel.GLOBAL && var instanceof BaseDataType[])
      {
         uninitializedGlobalExtents.get().add((BaseDataType[]) var);
      }
      
      return var;
   }
   
   /**
    * Checks if the new shared global extent variable was already initialized.
    * This method has the side-effect of removing the variable from the set of uninitialized
    * variables so any subsequent call in same session will return <code>false</code> excluding
    * this way the possible double initialization of the extent variable.
    * 
    * @param   var
    *          The global shared extent variable to be checked.
    * 
    * @return  <code>true</code> for the first call since the variable was added (declared),
    *          <code>false</code> for any subsequent calls for same variable in same context/
    *          session, or if the variable is not an extent.
    */
   public static boolean isInitialized(BaseDataType[] var)
   {
      return !uninitializedGlobalExtents.get().remove(var);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first
    * variable with the given name.
    *
    * @param    <T> 
    *           Variable type.
    * @param    name
    *           The variable name.
    *           
    * @return   The variable reference associated with the name or <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the variable cannot be found and silent error mode is disabled.
    */
   public static <T> T lookupVariable(String name)
   throws ErrorConditionException
   {
      return (T) lookupWorker(vars.obtainWA(), name, VAR_ERR_MSG, 392); 
   }
   
   /**
    * Adds a stream to the {@link ScopeLevel#CURRENT} scope.
    *
    * @param    name
    *           The stream name.
    * @param    stream
    *           The stream reference.
    */
   public static void addStream(String name, Stream stream)
   {
      addStream(ScopeLevel.CURRENT, name, stream);
   }
   
   /**
    * Adds a stream to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    *
    * @param    scope 
    *           Specifies the target scope.
    * @param    name
    *           The stream name.
    * @param    stream
    *           The stream reference.
    */
   public static void addStream(ScopeLevel scope, String name, Stream stream)
   {
      addWorker(streams.obtainWA(), scope, name, stream);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first
    * stream with the given name.
    *
    * @param    name
    *           The stream name.
    *           
    * @return   The stream reference associated with the name or
    *           <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the stream cannot be found and silent error mode is
    *           disabled.
    */
   public static Stream lookupStream(String name)
   throws ErrorConditionException
   {
      return (Stream) lookupWorker(streams.obtainWA(), name, STREAM_ERR_MSG, 567);
   }
   
   /**
    * Adds a buffer to the {@link ScopeLevel#CURRENT} scope.
    *
    * @param    name
    *           The buffer name.
    * @param    buffer
    *           The buffer reference.
    */
   public static <T> void addBuffer(String name, T buffer)
   {
      addBuffer(ScopeLevel.CURRENT, name, buffer);
   }
   
   /**
    * Adds a buffer to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    *
    * @param    <T>
    *           Buffer type.
    * @param    scope 
    *           Specifies the target scope.
    * @param    name
    *           The buffer name.
    * @param    buffer
    *           The buffer reference.
    */
   public static <T> void addBuffer(ScopeLevel scope, String name, T buffer)
   {
      addWorker(buffers.obtainWA(), scope, name, buffer);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first
    * buffer with the given name.
    *
    * @param   <T> 
    *          Buffer type.
    * @param   name
    *          The buffer name.
    *
    * @return  The buffer reference associated with the name or
    *          <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the buffer cannot be found and silent error mode is disabled.
    */
   public static <T> T lookupBuffer(String name)
   throws ErrorConditionException
   {
      return (T) lookupWorker(buffers.obtainWA(), name, BUF_ERR_MSG, 566);
   }
   
   /**
    * Adds a temporary table buffer to the {@link ScopeLevel#CURRENT} scope.
    *
    * @param   name
    *          The temporary table buffer name.
    * @param   temp
    *          The temporary table buffer reference.
    */
   public static void addTempTable(String name, Temporary temp)
   {
      addTempTable(ScopeLevel.CURRENT, name, temp);
   }
   
   /**
    * Adds a temporary table buffer to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is opened.
    * <p>
    * If global is specified, the instance will be added to the bottom-most
    * scope on the stack.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    *
    * @param   scope 
    *          Specifies the target scope.
    * @param   name
    *          The temporary table buffer name.
    * @param   temp
    *          The temporary table buffer reference.
    */
   public static void addTempTable(ScopeLevel scope, String name, Temporary temp)
   {
      addWorker(tempTables.obtainWA(), scope, name, temp);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first temp
    * table with the given name.
    *
    * @param   <T> 
    *          Variable type.
    * @param   name
    *          The temp table buffer name.
    *
    * @return  The temp table buffer reference associated with the name or
    *          <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the temp table cannot be found and silent error mode is
    *           disabled.  Note that the error message differs slightly from
    *           Progress, in that the procedure name is omitted.
    */
   public static <T extends Temporary> T lookupTempTable(String name)
   throws ErrorConditionException
   {
      return (T) lookupWorker(tempTables.obtainWA(), name, TT_ERR_MSG, 1429);
   }
   
   /**
    * Adds a frame to the {@link ScopeLevel#NEXT} scope.
    * <p>
    * The instance will be placed in the list of deferred instances that will
    * be added the next time a scope is opened.
    *
    * @param    name
    *           The frame name.
    * @param    frame
    *           The frame reference.
    *
    * @return   Previous instance with the same name or <code>null</code> if
    *           there were no such instance.  If there is a previous instance,
    *           it will be replaced with this new instance.
    */
   public static <T> T addFrame(String name, T frame)
   {
      return addFrame(ScopeLevel.NEXT, name, frame);
   }
   
   /**
    * Adds a frame to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is
    * opened.
    * <p>
    * If global is specified, the instance will be added to the next
    * scope on the stack, as there is no such thing as a global frame in
    * the Progress 4GL.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    *
    * @param    <T>
    *           Frame type.
    * @param    scope 
    *           Specifies the target scope.
    * @param    name
    *           The frame name.
    * @param    frame
    *           The frame reference.
    *
    * @return   Previous instance with the same name or <code>null</code> if
    *           there were no such instance.  If there is a previous instance,
    *           it will be replaced with this new instance.
    */
   public static <T> T addFrame(ScopeLevel scope, String name, T frame)
   {
      if (scope == ScopeLevel.GLOBAL)
      {
         scope = ScopeLevel.NEXT;
      }
      
      WorkArea<Object> wa = frames.obtainWA();
      T result = (T) wa.shared.lookupSymbol(name);
      
      addWorker(wa, scope, name, frame);
      
      return result;
   }
   
   /**
    * Adds a menu definition to the {@link ScopeLevel#NEXT} scope.
    * <p>
    * The instance will be placed in the list of deferred instances that will
    * be added the next time a scope is opened.
    *
    * @param    name
    *           The menu name.
    * @param    menu
    *           The menu definition reference.
    */
   public static <T> void addMenu(String name, T menu)
   {
      addMenu(ScopeLevel.NEXT, name, menu);
   }
   
   /**
    * Adds a menu to the specified scope.
    * <p>
    * If the next scope is specified, the instance will be placed in the list
    * of deferred instances that will be added the next time a scope is
    * opened.
    * <p>
    * If global is specified, the instance will be added to the next
    * scope on the stack, as GLOBAL is not a valid option for shared menus.
    * <p>
    * If current is specified, the instance will be added to the scope that
    * is on the top of the stack. If there is only one scope on the stack,
    * the global and current scope will be the same scope.
    *
    * @param    <T>
    *           Menu definition type.
    * @param    scope 
    *           Specifies the target scope. 
    * @param    name
    *           The menu name.
    * @param    menu
    *           The menu reference.
    */
   public static <T> void addMenu(ScopeLevel scope, String name, T menu)
   {
      if (scope == ScopeLevel.GLOBAL)
      {
         scope = ScopeLevel.NEXT;
      }
      
      WorkArea<Object> wa = menus.obtainWA();
      
      addWorker(wa, scope, name, menu);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first frame with the given name.
    *
    * @param   <T> 
    *          Frame type.
    * @param   name
    *          The frame name.
    *
    * @return  The frame reference associated with the name or <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the frame cannot be found and silent error mode is disabled.
    */
   public static <T> T lookupFrame(String name)
   throws ErrorConditionException
   {
      return (T) lookupWorker(frames.obtainWA(), name, FR_ERR_MSG, 682);
   }
   
   /**
    * Searches all scopes from the top of the stack down for the first menu with the given name.
    *
    * @param   <T> 
    *          menu definition type.
    * @param   name
    *          The menu name.
    *
    * @return  The menu definition reference associated with the name or <code>null</code> if not found.
    *
    * @throws   ErrorConditionException
    *           When the menu cannot be found and silent error mode is disabled.
    */
   public static <T> T lookupMenu(String name)
   throws ErrorConditionException
   {
      return (T) lookupWorker(menus.obtainWA(), name, FR_ERR_MSG, 682);
   }
   
   /**
    * Provides the singleton instance of this class as a {@link Scopeable}.
    *
    * @return   The only instance of this class.
    */
   public Scopeable createScopeable()
   {
      return this;
   }
   
   /**
    * Provides a notification that a new scope is about to be entered.
    */
   public void scopeStart()
   {
      addScope();
   }   
   
   /**
    * Provides a notification that a scope is about to be exited.
    */
   public void scopeFinished()
   {
      deleteScope();
   }
   
   /**
    * Provides notification that the external procedure scope has been deleted.
    * <p>
    * This is a no-op for the shared variable manager.
    */
   @Override
   public void scopeDeleted()
   {
      // no-op
   }
   
   /**
    * Core add worker that defers adding (using a pending list in the work area) if this is a 
    * <code>ScopeLevel.NEXT</code> request, otherwise the instance will be mapped into the current
    * or global scope immediately.
    *
    * @param    wa
    *           The context-local state.
    * @param    scope
    *           The scope to target.
    * @param    name
    *           The symbolic name for the shared instance.
    * @param    var
    *           The instance to be shared.
    */
   private static void addWorker(WorkArea<Object> wa, ScopeLevel scope, String name, Object var)
   {
      if (scope == ScopeLevel.NEXT)
      {
         if (wa.pending == null)
         {
            wa.pending = new HashMap<>();
         }
         
         String key = name;
         
         if (!wa.shared.isCaseSensitive())
         {
            key = name.toLowerCase();
         }
         
         // we don't have to worry about the global flag as these can't
         // be global, just save off the name and instance
         wa.pending.put(key, var);
      }
      else
      {
         wa.shared.addSymbol((scope == ScopeLevel.GLOBAL), name, var);
      }
   }
   
   /**
    * Core lookup worker that returns the first instance found by the
    * given name.  The order of lookup is:
    * <ol>
    *   <li> pending list
    *   <li> each of the scopes starting at the current scope and moving
    *        through the stack to the global scope (if it exists)
    * </ol>
    *
    * @param    wa
    *           The context-local state.
    * @param    name
    *           The symbolic name for the shared instance.
    * @param    errSpec
    *           If no instance is found, this is the error text which is
    *           combined with the resource name and the error number when
    *           the exception is thrown.
    * @param    errNum
    *           If no instance is found, this is the error number to be
    *           thrown.
    */
   private static <T> T lookupWorker(WorkArea<Object> wa,
                                     String           name,
                                     String           errSpec,
                                     int              errNum)
   throws ErrorConditionException
   {
      T result = null;
      
      if (wa.pending != null)
      {
         String key = name;
         
         if (!wa.shared.isCaseSensitive())
         {
            key = name.toLowerCase();
         }
         
         result = (T) wa.pending.get(key);
      }
      
      if (result == null)
      {
         result = (T) wa.shared.lookupSymbol(name);
      }
      
      if (result == null)
      {
         errorHelper(errNum, errSpec, name);
      }
      
      return result;
   }
   
   /**
    * Simple helper to record or throw the error needed using the {@link ErrorManager}.
    *
    * @param    num
    *           The error number.
    * @param    spec
    *           A <code>String.format</code> style specification string with
    *           only a single "%s" replacement string.
    * @param    name
    *           The resource name that was the cause of the error.
    *
    * @throws   ErrorConditionException
    *           When silent error mode is not enabled.
    */
   private static void errorHelper(int num, String spec, String name)
   throws ErrorConditionException
   {
      ErrorManager.recordOrThrowError(num, String.format(spec, name));
   }
   
   /** 
    * Stores the state of the current context.
    */
   private static class WorkArea<V>
   {
      /** Instances mapped by name and organized by scope. */
      private ScopedSymbolDictionary<V> shared = null;
      
      /** Instances that must be added to the NEXT scope. */
      private Map<String, V> pending = null;
   }
   
   /**
    * Simple <code>ContextLocal</code> subclass that properly initializes the
    * context-local reference to a new instance of a case-sensitive
    * <code>ScopedSymbolDictionary</code> that has a global scope.
    */
   private static class ContextContainer<V>
   extends ContextLocal<WorkArea>
   {
      /** Defines the case sensitivity of the contained dictionary. */ 
      private boolean caseSensitive = true;
      
      /**
       * Default constructor for a case-sensitive dictionary with a single global scope.
       */
      ContextContainer()
      {
         // nothing to do
      }
      
      /**
       * Constructor for a dictionary with a single global scope and
       * case-sensitivity as defined by the caller.
       *
       * @param   caseSensitive
       *          <code>true</code> for a case-sensitive dictionary, otherwise <code>false</code>.
       */
      ContextContainer(boolean caseSensitive)
      {
         this.caseSensitive = caseSensitive;
      }
      
      /**
       * Obtains the context-local instance of the contained dictionary.
       *
       * @return   The scoped dictionary associated with this context.
       */
      public ScopedSymbolDictionary<V> obtain() 
      {
         return this.get().shared;
      }
      
      /**
       * Obtains the context-local instance of the contained dictionary.
       *
       * @return   The scoped dictionary associated with this context.
       */
      public WorkArea<V> obtainWA() 
      {
         return this.get();
      }
      
      /**
       * Initializes the dictionary, the first time it is requested within a new context.
       *
       * @return   The newly instantiated dictionary.
       */
      protected synchronized WorkArea<V> initialValue()
      {
         // instantiates a dictionary with a global scope and which is case-sensitive
         WorkArea wa = new WorkArea<V>();
         wa.shared = new ScopedSymbolDictionary<V>(null, caseSensitive);
         
         return wa;
      }
   }
}
